@@ -186,10 +186,11 @@ class LoopIterator(object):
 class Statement(object):
     "One statement in a QMASM source file."
 
-    def __init__(self, qmasm, filename, lineno):
+    def __init__(self, qmasm, filename, lineno, as_qubo):
         self.qmasm = qmasm
         self.filename = filename
         self.lineno = lineno
+        self.as_qubo = as_qubo
 
     def error_in_line(self, msg):
         if self.lineno == None:
@@ -208,8 +209,8 @@ class Statement(object):
 
 class Weight(Statement):
     "Represent a point weight on a qubit."
-    def __init__(self, qmasm, filename, lineno, sym, weight):
-        super(Weight, self).__init__(qmasm, filename, lineno)
+    def __init__(self, qmasm, filename, lineno, as_qubo, sym, weight):
+        super(Weight, self).__init__(qmasm, filename, lineno, as_qubo)
         self.sym = self.validate_ident(sym)
         self.weight = weight
 
@@ -218,12 +219,15 @@ class Weight(Statement):
 
     def update_qmi(self, prefix, next_prefix, problem):
         num = self.qmasm.symbol_to_number(prefix + self.sym, prefix, next_prefix)
-        problem.weights[num] += self.weight
+        if self.as_qubo:
+            problem.weights[num] += self.weight/2.0
+        else:
+            problem.weights[num] += self.weight
 
 class Chain(Statement):
     "Chain between qubits."
-    def __init__(self, qmasm, filename, lineno, sym1, sym2):
-        super(Chain, self).__init__(qmasm, filename, lineno)
+    def __init__(self, qmasm, filename, lineno, as_qubo, sym1, sym2):
+        super(Chain, self).__init__(qmasm, filename, lineno, as_qubo)
         self.sym1 = self.validate_ident(sym1)
         self.sym2 = self.validate_ident(sym2)
 
@@ -244,8 +248,8 @@ class Chain(Statement):
 
 class AntiChain(Statement):
     "AntiChain between qubits."
-    def __init__(self, qmasm, filename, lineno, sym1, sym2):
-        super(AntiChain, self).__init__(qmasm, filename, lineno)
+    def __init__(self, qmasm, filename, lineno, as_qubo, sym1, sym2):
+        super(AntiChain, self).__init__(qmasm, filename, lineno, as_qubo)
         self.sym1 = self.validate_ident(sym1)
         self.sym2 = self.validate_ident(sym2)
 
@@ -266,8 +270,8 @@ class AntiChain(Statement):
 
 class Pin(Statement):
     "Pinning of a qubit to true or false."
-    def __init__(self, qmasm, filename, lineno, sym, goal):
-        super(Pin, self).__init__(qmasm, filename, lineno)
+    def __init__(self, qmasm, filename, lineno, as_qubo, sym, goal):
+        super(Pin, self).__init__(qmasm, filename, lineno, as_qubo)
         self.sym = self.validate_ident(sym)
         self.goal = goal
 
@@ -281,9 +285,9 @@ class Pin(Statement):
         problem.pending_asserts.append((sym, "=", str(int(self.goal))))
 
 class Alias(Statement):
-    "Alias of one symbol to another."
-    def __init__(self, qmasm, filename, lineno, sym1, sym2):
-        super(Alias, self).__init__(qmasm, filename, lineno)
+    "Alias one symbol to another."
+    def __init__(self, qmasm, filename, lineno, as_qubo, sym1, sym2):
+        super(Alias, self).__init__(qmasm, filename, lineno, as_qubo)
         self.sym1 = self.validate_ident(sym1)
         self.sym2 = self.validate_ident(sym2)
 
@@ -298,10 +302,24 @@ class Alias(Statement):
             sym2 = sym2.replace(prefix + "!next.", next_prefix)
         self.qmasm.sym_map.alias(sym1, sym2)
 
+class BQMType(Statement):
+    "Set the BQM mode to either Ising or QUBO."
+    def __init__(self, qmasm, filename, lineno, as_qubo):
+        super(BQMType, self).__init__(qmasm, filename, lineno, as_qubo)
+
+    def as_str(self, prefix=""):
+        if self.as_qubo:
+            return "!bqm_type qubo"
+        else:
+            return "!bqm_type ising"
+
+    def update_qmi(self, prefix, next_prefix, problem):
+        pass
+
 class Rename(Statement):
     "Rename one set of symbols to another."
-    def __init__(self, qmasm, filename, lineno, syms1, syms2):
-        super(Rename, self).__init__(qmasm, filename, lineno)
+    def __init__(self, qmasm, filename, lineno, as_qubo, syms1, syms2):
+        super(Rename, self).__init__(qmasm, filename, lineno, as_qubo)
         self.syms1 = [self.validate_ident(s) for s in syms1]
         self.syms2 = [self.validate_ident(s) for s in syms2]
 
@@ -401,8 +419,8 @@ class Rename(Statement):
 
 class Strength(Statement):
     "Coupler strength between two qubits."
-    def __init__(self, qmasm, filename, lineno, sym1, sym2, strength):
-        super(Strength, self).__init__(qmasm, filename, lineno)
+    def __init__(self, qmasm, filename, lineno, as_qubo, sym1, sym2, strength):
+        super(Strength, self).__init__(qmasm, filename, lineno, as_qubo)
         self.sym1 = self.validate_ident(sym1)
         self.sym2 = self.validate_ident(sym2)
         self.strength = strength
@@ -417,13 +435,19 @@ class Strength(Statement):
             self.error_in_line("A coupler cannot connect a spin to itself")
         elif num1 > num2:
             num1, num2 = num2, num1
-        problem.strengths[(num1, num2)] += self.strength
+        if self.as_qubo:
+            s4 = self.strength/4.0
+            problem.strengths[(num1, num2)] += s4
+            problem.weights[num1] += s4
+            problem.weights[num2] += s4
+        else:
+            problem.strengths[(num1, num2)] += self.strength
 
 class Assert(Statement):
     "Instantiation of a run-time assertion."
 
-    def __init__(self, qmasm, filename, lineno, expr):
-        super(Assert, self).__init__(qmasm, filename, lineno)
+    def __init__(self, qmasm, filename, lineno, as_qubo, expr):
+        super(Assert, self).__init__(qmasm, filename, lineno, as_qubo)
         self.parser = AssertParser(qmasm)
         self.expr = expr
         self.ast = self.parser.parse(expr)
@@ -449,8 +473,8 @@ class Assert(Statement):
 
 class MacroUse(Statement):
     "Instantiation of a macro definition."
-    def __init__(self, qmasm, filename, lineno, name, body, prefixes):
-        super(MacroUse, self).__init__(qmasm, filename, lineno)
+    def __init__(self, qmasm, filename, lineno, as_qubo, name, body, prefixes):
+        super(MacroUse, self).__init__(qmasm, filename, lineno, as_qubo)
         self.name = self.validate_ident(name)
         self.body = body
         self.prefixes = [self.validate_ident(p) + "." for p in prefixes]
@@ -466,7 +490,7 @@ class MacroUse(Statement):
                 if "!next." not in sstr:
                     stmt_strs.append(sstr)
         else:
-            # At least one prefix -- disaply the macro body in a new scope
+            # At least one prefix -- display the macro body in a new scope
             # (i.e., by augmenting the given prefix with each new prefix in
             # turn).
             for p in range(nprefixes):
@@ -520,6 +544,7 @@ class FileParser(object):
         self.env = Environment()      # Stack of maps from compile-time variable names to values
         self.expr_parser = ExprParser(qmasm)     # Expression parser
         self.rel_parser = RelationParser(qmasm)  # Relation parser
+        self._as_qubo = False   # Current BQM mode (QUBO or Ising)
 
         # Establish a mapping from a first-field directive to a parsing function.
         self.dir_to_func = {
@@ -529,7 +554,8 @@ class FileParser(object):
             "!begin_macro": self.parse_line_begin_macro,
             "!end_macro":   self.parse_line_end_macro,
             "!use_macro":   self.parse_line_use_macro,
-            "!alias":       self.parse_line_sym_alias
+            "!alias":       self.parse_line_sym_alias,
+            "!bqm_type":    self.parse_line_bqm_type
         }
 
     def is_float(self, str):
@@ -596,7 +622,7 @@ class FileParser(object):
         # "!assert" <expr> -- assert a property that must be true at run time.
         if len(fields) < 2:
             error_in_line(filename, lineno, "Expected an expression to follow !assert")
-        self.target.append(Assert(self.qmasm, filename, lineno, " ".join(self.env.sub_syms(fields[1:]))))
+        self.target.append(Assert(self.qmasm, filename, lineno, self._as_qubo, " ".join(self.env.sub_syms(fields[1:]))))
 
     def parse_line_let(self, filename, lineno, fields):
         "Parse a !let directive."
@@ -653,7 +679,7 @@ class FileParser(object):
             val = float(self.env.sub_syms(fields[1]))
         except ValueError:
             error_in_line(filename, lineno, 'Failed to parse "%s %s" as a symbol followed by a numerical weight' % (fields[0], fields[1]))
-        self.target.append(Weight(self.qmasm, filename, lineno, self.env.sub_syms(fields[0]), val))
+        self.target.append(Weight(self.qmasm, filename, lineno, self._as_qubo, self.env.sub_syms(fields[0]), val))
 
     def parse_line_chain(self, filename, lineno, fields):
         "Parse a qubit chain."
@@ -711,7 +737,7 @@ class FileParser(object):
         if num_arrows != 1 or rhs[0] != "->":
             error_in_line(filename, lineno, 'Failed to parse "%s" as a symbol rename' % (" ".join(fields)))
         rhs = rhs[1:]  # Drop the "->".
-        self.target.append(Rename(self.qmasm, filename, lineno, lhs, rhs))
+        self.target.append(Rename(self.qmasm, filename, lineno, self._as_qubo, lhs, rhs))
 
     def parse_line_strength(self, filename, lineno, fields):
         "Parse a coupler strength."
@@ -722,7 +748,7 @@ class FileParser(object):
             strength = float(self.env.sub_syms(fields[2]))
         except ValueError:
             error_in_line(filename, lineno, 'Failed to parse "%s" as a number' % fields[2])
-        self.target.append(Strength(self.qmasm, filename, lineno, self.env.sub_syms(fields[0]), self.env.sub_syms(fields[1]), strength))
+        self.target.append(Strength(self.qmasm, filename, lineno, self._as_qubo, self.env.sub_syms(fields[0]), self.env.sub_syms(fields[1]), strength))
 
     def parse_line_use_macro(self, filename, lineno, fields):
         "Parse a !use_macro directive."
@@ -733,7 +759,7 @@ class FileParser(object):
         name = self.env.sub_syms(fields[1])
         prefixes = [self.env.sub_syms(p) for p in fields[2:]]
         try:
-            self.target.append(MacroUse(self.qmasm, filename, lineno, name, self.macros[name], prefixes))
+            self.target.append(MacroUse(self.qmasm, filename, lineno, self._as_qubo, name, self.macros[name], prefixes))
         except KeyError:
             error_in_line(filename, lineno, "Unknown macro %s" % name)
 
@@ -744,6 +770,18 @@ class FileParser(object):
             error_in_line(filename, lineno, "Expected a symbol name and replacement to follow !alias")
         self.env[fields[1]] = self.env.sub_syms(fields[2])
 
+    def parse_line_bqm_type(self, filename, lineno, fields):
+        "Parse a !bqm_type directive."
+        if len(fields) != 2:
+            error_in_line(filename, lineno, 'Expected either "qubo" or "ising" to follow !bqm_type')
+        if fields[1] == "qubo":
+            as_qubo = True
+        elif fields[1] == "ising":
+            as_qubo = False
+        else:
+            error_in_line(filename, lineno, 'Expected either "qubo" or "ising" to follow !bqm_type')
+        self.target.append(BQMType(self.qmasm, filename, lineno, as_qubo))
+
     def process_if(self, filename, lineno, fields, all_lines):
         """Parse and process an !if directive.  Recursively parse the remaining
         file contents."""
@@ -867,7 +905,7 @@ class FileParser(object):
         rhs_list = pin_parser.parse_rhs(self.env.sub_syms(lhs_rhs[1]))
         if len(lhs_list) != len(rhs_list):
             self.qmasm.abend('Different number of left- and right-hand-side values in "%s" (%d vs. %d)' % (pin_str, len(lhs_list), len(rhs_list)))
-        return [Pin(self.qmasm, filename, lineno, l, r) for l, r in zip(lhs_list, rhs_list)]
+        return [Pin(self.qmasm, filename, lineno, self._as_qubo, l, r) for l, r in zip(lhs_list, rhs_list)]
 
     def process_chain(self, filename, lineno, chain_str):
         "Parse a chain statement into one or more Chain objects and add these to the program."
@@ -880,7 +918,7 @@ class FileParser(object):
         rhs_list = pin_parser.parse_lhs(self.env.sub_syms(lhs_rhs[1]))  # Note use of parse_lhs to parse the RHS.
         if len(lhs_list) != len(rhs_list):
             self.qmasm.abend('Different number of left- and right-hand-side values in "%s" (%d vs. %d)' % (chain_str, len(lhs_list), len(rhs_list)))
-        return [Chain(self.qmasm, filename, lineno, l, r) for l, r in zip(lhs_list, rhs_list)]
+        return [Chain(self.qmasm, filename, lineno, self._as_qubo, l, r) for l, r in zip(lhs_list, rhs_list)]
 
     def process_antichain(self, filename, lineno, antichain_str):
         "Parse an anti-chain statement into one or more AntiChain objects and add these to the program."
@@ -893,7 +931,7 @@ class FileParser(object):
         rhs_list = pin_parser.parse_lhs(self.env.sub_syms(lhs_rhs[1]))  # Note use of parse_lhs to parse the RHS.
         if len(lhs_list) != len(rhs_list):
             self.qmasm.abend('Different number of left- and right-hand-side values in "%s" (%d vs. %d)' % (antichain_str, len(lhs_list), len(rhs_list)))
-        return [AntiChain(self.qmasm, filename, lineno, l, r) for l, r in zip(lhs_list, rhs_list)]
+        return [AntiChain(self.qmasm, filename, lineno, self._as_qubo, l, r) for l, r in zip(lhs_list, rhs_list)]
 
     def process_alias(self, filename, lineno, alias_str):
         "Parse an alias statement into one or more Alias objects and add these to the program."
@@ -906,7 +944,7 @@ class FileParser(object):
         rhs_list = pin_parser.parse_lhs(self.env.sub_syms(lhs_rhs[1]))  # Note use of parse_lhs to parse the RHS.
         if len(lhs_list) != len(rhs_list):
             self.qmasm.abend('Different number of left- and right-hand-side values in "%s" (%d vs. %d)' % (alias_str, len(lhs_list), len(rhs_list)))
-        return [Alias(self.qmasm, filename, lineno, l, r) for l, r in zip(lhs_list, rhs_list)]
+        return [Alias(self.qmasm, filename, lineno, self._as_qubo, l, r) for l, r in zip(lhs_list, rhs_list)]
 
     def process_file_contents(self, filename, all_lines):
         """Parse the contents of a file.  Contents are passed as a list plus an
@@ -966,6 +1004,8 @@ class FileParser(object):
                     # None of the above
                     error_in_line(filename, lineno, 'Failed to parse "%s"' % line)
             func(filename, lineno, fields)
+            if func == self.parse_line_bqm_type:
+                self._as_qubo = self.target[-1].as_qubo
 
     def process_file(self, filename, infile):
         """Define a function that parses an input file into an internal
