@@ -5,6 +5,7 @@
 
 import copy
 import dimod
+import re
 import sys
 from collections import defaultdict
 from qmasm.assertions import AssertParser
@@ -13,11 +14,10 @@ from qmasm.assertions import AssertParser
 # it turns out we don't even need this veneer, we may replace it with direct
 # use of dimod.BinaryQuadraticModel.
 class Problem(object):
-    "Represent either an Ising or QUBO problem."
+    "Represent an Ising problem."
 
-    def __init__(self, qmasm, qubo):
+    def __init__(self, qmasm):
         self.qmasm = qmasm   # Pointer to the top-level QMASM class
-        self.qubo = qubo     # True=QUBO; False=Ising
         self.weights = defaultdict(lambda: 0.0)    # Map from a spin to a point weight
         self.strengths = defaultdict(lambda: 0.0)  # Map from a pair of spins to a coupler strength
         self.chains = set()       # Subset of strengths keys that represents user-defined chains (always logical)
@@ -45,10 +45,6 @@ class Problem(object):
                 except ValueError:
                     # No weights or strengths -- arbitrarily choose -1.
                     chain_strength = -1.0
-        elif self.qubo:
-            # With QUBO input we need to divide the chain strength by 4 for
-            # consistency with the other coupler strengths.
-            chain_strength /= 4.0
         for c in self.chains:
             self.strengths[c] += chain_strength
         for c in self.antichains:
@@ -57,13 +53,8 @@ class Problem(object):
 
     def generate_bqm(self):
         "Generate a BinaryQuadraticModel version of the Problem."
-        # Create a BQM.
-        btype = dimod.SPIN
-        if self.qubo:
-            btype = dimod.BINARY
-        bqm = dimod.BinaryQuadraticModel(self.weights, self.strengths, 0, btype, problem=self)
-        if self.qubo:
-            bqm.change_vartype(dimod.SPIN, inplace=True)
+        # Create an Ising-model BQM.
+        bqm = dimod.BinaryQuadraticModel(self.weights, self.strengths, 0, dimod.SPIN)
 
         # Pin all variables the user asked to pin.
         bool2spin = {False: -1, True: +1}
@@ -103,6 +94,17 @@ class Problem(object):
             self.bqm_vars.add(v)
         return self.bqm_vars
 
+    def numeric_variables(self):
+        """Return a mapping from a variable name to a list of physical qubits
+        to which it should be mapped."""
+        var2phys = {}
+        num_re = re.compile(r'\d+')
+        for s, n in self.qmasm.sym_map.symbol_number_items():
+            nums = num_re.findall(s)
+            if nums != []:
+                var2phys[n] = [int(k) for k in nums]
+        return var2phys
+
     def _edges_to_adj_list(self, edges):
         "Convert a list of edges to an adjacency list."
         adj = defaultdict(lambda: [])
@@ -251,71 +253,116 @@ class Problem(object):
         except KeyError:
             pass  # Probably a software solver.
 
-    def _output_embedding_verbosely(self, max_sym_name_len, num2syms):
-        "Verbosely output the mapping from logical to physical qubits."
-        sys.stderr.write("Established a mapping from logical to physical qubits:\n\n")
-        sys.stderr.write("    Logical  %-*s  Physical\n" % (max_sym_name_len, "Name(s)"))
-        sys.stderr.write("    -------  %s  -------------------------\n" % ("-" * max_sym_name_len))
+    def _describe_logical_to_physical(self, num2syms, reduce_contractions):
+        "Map each logical qubit to zero or more physical qubits."
+        # On the first pass, assign a descriptive tuple to each logical qubit
+        # number.
+        log2phys = []   # Map from a logical qubit to a descriptive tuple
         known_values = self.logical.merged_known_values()
         pin_map = {k: v for k, v in self.logical.pinned}
         for i in range(len(num2syms)):
             if num2syms[i] == []:
+                log2phys.append(("unused",))
                 continue
-            name_list = " ".join(sorted(num2syms[i]))
             try:
-                phys_list = " ".join(["%4d" % e for e in sorted(self.embedding[i])])
+                log2phys.append(("physical", sorted(self.embedding[i])))
             except KeyError:
                 try:
-                    phys_list = "[Pinned to %s]" % repr(pin_map[i])
+                    log2phys.append(("pinned", pin_map[i]))
                 except KeyError:
                     try:
-                        phys_list = "[Provably %s]" % known_values[i]
+                        log2phys.append(("known", known_values[i]))
                     except KeyError:
                         try:
-                            phys_list = "[Same as logical %d]" % self.contractions[i]
+                            log2phys.append(("same_as", self.contractions[i]))
                         except KeyError:
                             if self.embedding == {}:
-                                phys_list = str(i)
+                                log2phys.append(("physical", [i]))
                             else:
-                                phys_list = "[Disconnected]"
-            sys.stderr.write("    %7d  %-*s  %s\n" % (i, max_sym_name_len, name_list, phys_list))
+                                log2phys.append(("disconnected",))
+
+        # On the second, optional, pass, replace same_as tuples with whatever
+        # they transitively point to.
+        if reduce_contractions:
+            for i in range(len(log2phys)):
+                while log2phys[i][0] == "same_as":
+                    log2phys[i] = log2phys[log2phys[i][1]]
+        return log2phys
+
+    def _output_embedding_verbosely(self, max_sym_name_len, num2syms, verbosity):
+        "Verbosely output the mapping from logical to physical qubits."
+        sys.stderr.write("Established a mapping from logical to physical qubits:\n\n")
+        sys.stderr.write("    Logical  %-*s  Physical\n" % (max_sym_name_len, "Name(s)"))
+        sys.stderr.write("    -------  %s  -------------------------\n" % ("-" * max_sym_name_len))
+        log2phys_desc = self._describe_logical_to_physical(num2syms, verbosity < 2)
+        for i in range(len(log2phys_desc)):
+            # Ignore unused logical qubits and "uninteresting" symbols (i.e.,
+            # those containing a "$").
+            desc = log2phys_desc[i]
+            tag = desc[0]
+            if tag == "unused":
+                continue
+            syms = num2syms[i]
+            if verbosity < 2:
+                syms = [s for s in syms if "$" not in s]
+            if len(syms) == 0:
+                continue
+            name_list = " ".join(sorted(syms))
+
+            # Determine how to display the qubit based on the tag.
+            if tag == "physical":
+                phys_str = " ".join(["%4d" % e for e in desc[1]])
+            elif tag == "pinned":
+                phys_str = "[Pinned to %s]" % repr(desc[1])
+            elif tag == "known":
+                phys_str = "[Provably %s]" % repr(desc[1])
+            elif tag == "same_as":
+                phys_str = "[Same as logical %d]" % desc[1]
+            elif tag == "disconnected":
+                phys_str = "[Disconnected]"
+            else:
+                self.qmasm.abend('Internal error: Unknown tag type "%s"' % tag)
+            sys.stderr.write("    %7d  %-*s  %s\n" % (i, max_sym_name_len, name_list, phys_str))
         sys.stderr.write("\n")
 
     def _output_embedding_tersely(self, max_sym_name_len, num2syms):
         "Tersely output the mapping from logical to physical qubits."
         log2phys_comments = []
-        known_values = self.logical.merged_known_values()
-        pin_map = {k: v for k, v in self.logical.pinned}
-        for i in range(len(num2syms)):
-            if num2syms[i] == []:
+        log2phys_desc = self._describe_logical_to_physical(num2syms, True)
+        for i in range(len(log2phys_desc)):
+            # Ignore unused logical qubits and "uninteresting" symbols (i.e.,
+            # those containing a "$").
+            desc = log2phys_desc[i]
+            tag = desc[0]
+            if tag == "unused":
                 continue
-            name_list = " ".join(sorted(num2syms[i]))
-            try:
-                phys_list = " ".join(["%4d" % e for e in sorted(self.embedding[i])])
-            except KeyError:
-                try:
-                    phys_list = "[%s]" % repr(pin_map[i])
-                except KeyError:
-                    try:
-                        phys_list = "[%s]" % known_values[i]
-                    except KeyError:
-                        try:
-                            same = self.contractions[i]
-                            same_str = " ".join(self.qmasm.sym_map.to_symbols(same))
-                            phys_list = "[Same as %s]" % same_str
-                        except KeyError:
-                            if self.embedding == {}:
-                                phys_list = str(i)
-                            else:
-                                phys_list = "[Disconnected]"
-            log2phys_comments.append("# %s --> %s" % (name_list, phys_list))
+            syms = [s for s in num2syms[i] if "$" not in s]
+            if len(syms) == 0:
+                continue
+            name_list = " ".join(sorted(syms))
+
+            # Determine how to display the qubit based on the tag.
+            if tag == "physical":
+                phys_str = " ".join(["%4d" % e for e in desc[1]])
+            elif tag == "pinned":
+                phys_str = "[%s]" % repr(desc[1])
+            elif tag == "known":
+                phys_str = "[%s]" % repr(desc[1])
+            elif tag == "same_as":
+                same_str = " ".join(self.qmasm.sym_map.to_symbols(desc[1]))
+                phys_str = "[Same as logical %s]" % same_str
+            elif tag == "disconnected":
+                phys_str = "[Disconnected]"
+            else:
+                self.qmasm.abend('Internal error: Unknown tag type "%s"' % tag)
+            log2phys_comments.append("# %s --> %s" % (name_list, phys_str))
         log2phys_comments.sort()
         sys.stderr.write("\n".join(log2phys_comments) + "\n")
 
     def output_embedding(self, verbosity, max_sym_name_len, num2syms):
         "Output the mapping from logical to physical qubits."
         if verbosity > 0:
-            self._output_embedding_verbosely(max_sym_name_len, num2syms)
+            self._output_embedding_verbosely(max_sym_name_len, num2syms, verbosity)
         else:
             self._output_embedding_tersely(max_sym_name_len, num2syms)
 
