@@ -9,24 +9,31 @@ import hashlib
 import json
 import marshal
 import minorminer
+import networkx as nx
 import os
+import random
 import re
 import sys
+import threading
 import time
+from collections import defaultdict
 from concurrent.futures import ThreadPoolExecutor
 from dimod import ExactSolver, SampleSet
-from dwave.cloud import Client
+from dwave.cloud import Client, hybrid
+from dwave.cloud.exceptions import SolverFailureError
 from dwave.embedding import embed_bqm
-from dwave.system import DWaveSampler, EmbeddingComposite
+from dwave.system import DWaveSampler, EmbeddingComposite, VirtualGraphComposite, LeapHybridSampler
 from dwave_qbsolv import QBSolv
+from hybrid.reference.kerberos import KerberosSampler
 from neal import SimulatedAnnealingSampler
 from qmasm.solutions import Solutions
+from qmasm.utils import SpecializedPriorityQueue
 from tabu import TabuSampler
 
 class EmbeddingCache(object):
     "Read and write an embedding cache file."
 
-    def __init__(self, qmasm, edges, adj):
+    def __init__(self, qmasm, edges, adj, max_qubits, var2phys):
         # Ensure we have a valid cache directory.
         self.hash = None
         try:
@@ -41,6 +48,8 @@ class EmbeddingCache(object):
         sha = hashlib.sha1()
         sha.update(str(sorted(edges)).encode("utf-8"))
         sha.update(str(sorted(adj)).encode("utf-8"))
+        sha.update(str(max_qubits).encode("utf-8"))
+        sha.update(str(sorted(var2phys.items())).encode("utf-8"))
         self.hash = sha.hexdigest()
 
     def read(self):
@@ -71,66 +80,100 @@ class Sampler(object):
 
     # Keep track of parameters rejected by the solver.
     unexp_arg_re = re.compile(r"got an unexpected keyword argument '(\w+)'")
-    rejected_params = []
+    not_param_re = re.compile(r"(\w+) is not a parameter of this solver")
+    unk_param_re = re.compile(r"Unknown parameter (\w+)")
 
-    def __init__(self, qmasm, profile=None, solver=None, qbsolv=False):
+    def __init__(self, qmasm, profile=None, solver=None):
         "Acquire either a software sampler or a sampler representing a hardware solver."
         self.qmasm = qmasm
-        self.sampler, self.client_info = self._get_sampler(profile, solver)
-        if qbsolv:
-            if solver == None:
-                self.qbsolv_sampler = EmbeddingComposite(self.sampler)
-            else:
-                self.qbsolv_sampler = self.sampler
-            self.sampler = QBSolv()
-            self.client_info["solver_name"] = "QBSolv + " + self.client_info["solver_name"]
-        else:
-            self.qbsolv_sampler = None
+        self.profile = profile
+        self.sampler, self.client_info, self.extra_solver_params = self.get_sampler(profile, solver)
+        self.rejected_params = []
+        self.rejected_params_lock = threading.Lock()
+        self.final_params = {}   # Final parameters associated with the first QMI
+        self.final_params_sem = threading.Semaphore(0)
 
-    def _get_sampler(self, profile, solver):
-        "Return a dimod.Sampler object."
+    def get_sampler(self, profile, solver):
+        "Return a dimod.Sampler object and associated solver information."
         # Handle built-in software samplers as special cases.
         info = {}
         if solver != None:
             info["solver_name"] = solver
         if solver == "exact":
-            return ExactSolver(), info
+            return ExactSolver(), info, {}
         elif solver == "neal":
-            return SimulatedAnnealingSampler(), info
+            return SimulatedAnnealingSampler(), info, {}
         elif solver == "tabu":
-            return TabuSampler(), info
+            return TabuSampler(), info, {}
+        elif solver == "kerberos" or (solver != None and solver[:9] == "kerberos,"):
+            base_sampler = KerberosSampler()
+            try:
+                sub_sampler_name = solver.split(",")[1]
+            except IndexError:
+                sub_sampler_name = None
+            sub_sampler, sub_info, params = self.get_sampler_from_config(profile, sub_sampler_name, "qpu")
+            info.update(self._recursive_properties(sub_sampler))
+            info["solver_name"] = "kerberos + %s" % sub_info["solver_name"]
+            params["qpu_sampler"] = sub_sampler
+            return base_sampler, info, params
+        elif solver == "qbsolv" or (solver != None and solver[:7] == "qbsolv,"):
+            base_sampler = QBSolv()
+            try:
+                sub_sampler_name = solver.split(",")[1]
+            except IndexError:
+                sub_sampler_name = None
+            sub_sampler, sub_info, params = self.get_sampler(profile, sub_sampler_name)
+            if getattr(sub_sampler, "structure", None) != None:
+                sub_sampler = EmbeddingComposite(sub_sampler)
+            info.update(self._recursive_properties(sub_sampler))
+            info["solver_name"] = "QBSolv + %s" % sub_info["solver_name"]
+            params["solver"] = sub_sampler
+            return base_sampler, info, params
 
         # In the common case, read the configuration file, either the
         # default or the one named by the DWAVE_CONFIG_FILE environment
         # variable.
-        if profile != None:
-            info["profile"] = profile
+        return self.get_sampler_from_config(profile, solver)
+
+    def get_sampler_from_config(self, profile=None, solver=None, sampler_type=None):
+        """Return a dimod.Sampler object found in the user's configuration file,
+        associated solver information, and any extra parameters needed."""
         try:
-            with Client.from_config(profile=profile) as client:
+            with Client.from_config(profile=profile, client=sampler_type) as client:
                 if solver == None:
                     solver = client.default_solver
                 else:
                     solver = {"name": solver}
-                sampler = DWaveSampler(profile=profile, solver=solver)
-                info = {"solver_name": sampler.solver.name,
-                        "endpoint": client.endpoint}
-                return sampler, info
+                if isinstance(client, hybrid.Client):
+                    sampler = LeapHybridSampler(profile=profile, solver=solver)
+                else:
+                    sampler = DWaveSampler(profile=profile, solver=solver)
+                info = self._recursive_properties(sampler)
+                info["solver_name"] = sampler.solver.name
+                info["endpoint"] = client.endpoint
+                if profile != None:
+                    info["profile"] = profile
+                return sampler, info, {}
         except Exception as err:
             self.qmasm.abend("Failed to construct a sampler (%s)" % str(err))
 
+    def _recursive_properties(self, sampler):
+        "Perform a postfix traversal of a sampler's children's properties."
+        if sampler == None:
+            return {}
+        props = {}
+        for c in getattr(sampler, "children", []):
+            props.update(c.properties)
+        props.update(sampler.properties)
+        return props
+
     def show_properties(self, verbose):
         "Output either short or all solver properties."
         if verbose == 0:
             return
 
-        # Combine properties from various sources into a single dictionary.
-        props = self.client_info.copy()
-        if self.qbsolv_sampler != None:
-            props.update(self.qbsolv_sampler.properties)
-        else:
-            props.update(self.sampler.properties)
-
         # Determine the width of the widest key.
+        props = self.client_info.copy()
         max_key_len = len("Parameter")
         prop_keys = sorted(props)
         for k in prop_keys:
@@ -227,7 +270,7 @@ class Sampler(object):
                 adj_dict[v] = [u]
         return adj_dict
 
-    def _find_embedding(self, edges, adj, **kwargs):
+    def _find_embedding(self, edges, adj, max_qubits, **kwargs):
         "Wrap minorminer.find_embedding with a version that intercepts its output."
         # Minorminer accepts the hardware adjacency as a list of
         # pairs, not a map from each node to its neighbors.
@@ -274,7 +317,52 @@ class Sampler(object):
             embedding = json.loads(pipe.readline())
             return {int(k): v for k, v in embedding.items()}
 
-    def find_problem_embedding(self, logical, topology_file, verbosity):
+    def _adj_subset(self, hw_adj, num):
+        "Return a subset of a given size of the hardware adjacency graph."
+        # Construct a NetworkX graph from the given adjacency list.
+        # Set each node's popularity to 0.
+        G = nx.Graph()
+        for n1, ns in hw_adj.items():
+            G.add_edges_from([(n1, n2) for n2 in ns])
+        s0 = min(G.nodes)
+
+        # Construct a priority queue of nodes.
+        shuffled_nodes = list(G.nodes)
+        random.shuffle(shuffled_nodes)
+        priQ = SpecializedPriorityQueue(shuffled_nodes)
+
+        # Traverse the graph in order of decreasing node popularity.
+        priQ.increase_priority(s0)
+        visited = set()
+        order = []
+        while len(order) < num:
+            # Pop the most popular node.
+            node = priQ.pop_max()
+            if node in visited:
+                continue
+            visited.add(node)
+            order.append(node)
+
+            # Increase the popularity of the current node's immediate neighbors.
+            for nn in G.neighbors(node):
+                if nn not in visited:
+                    priQ.increase_priority(nn)
+
+        # Find all edges connecting the first num nodes, and convert these to
+        # the format Ocean expects.
+        adj = defaultdict(lambda: [])
+        for u, v in G.subgraph(order).edges:
+            adj[u].append(v)
+            adj[v].append(u)
+        return adj
+
+    def _impose_qubit_packing(self, max_qubits, edges, hw_adj, **embed_args):
+        'Pack qubits into a "corner" of the physical topology.'
+        packed_adj = self._adj_subset(hw_adj, max_qubits)
+        packed_embedding = self._find_embedding(edges, packed_adj, max_qubits, **embed_args)
+        return packed_embedding
+
+    def find_problem_embedding(self, logical, topology_file, max_qubits, physical_nums, verbosity):
         """Find an embedding of a problem on a physical topology, if
         necessary.  Return a physical Sampler object."""
         # Create a physical Problem.
@@ -289,6 +377,11 @@ class Sampler(object):
             hw_adj = self.read_hardware_adjacency(topology_file, verbosity)
         physical.hw_adj = hw_adj
 
+        # Identify any forced variable-to-qubit mappings.
+        forced_mappings = {}
+        if physical_nums:
+            forced_mappings = physical.numeric_variables()
+
         # See if we already have an embedding in the embedding cache.
         edges = logical.bqm.quadratic
         if hw_adj == None or len(edges) == 0:
@@ -297,7 +390,7 @@ class Sampler(object):
             return physical
         if verbosity >= 2:
             sys.stderr.write("Minor-embedding the logical problem onto the physical topology:\n\n")
-        ec = EmbeddingCache(self.qmasm, edges, hw_adj)
+        ec = EmbeddingCache(self.qmasm, edges, hw_adj, max_qubits, forced_mappings)
         if verbosity >= 2:
             if ec.cachedir == None:
                 sys.stderr.write("  No embedding cache directory was specified ($QMASMCACHE).\n")
@@ -318,12 +411,22 @@ class Sampler(object):
             sys.stderr.write("  No existing embedding found in the embedding cache.\n")
 
         # Minor-embed the logical problem onto the hardware topology.
-        if verbosity < 2:
-            physical.embedding = self._find_embedding(edges, hw_adj)
+        embed_args = {"tries": 100,
+                      "max_no_improvement": 25,
+                      "fixed_chains": forced_mappings,
+                      "verbose": max(verbosity - 1, 0)}
+        if max_qubits == None:
+            if verbosity >= 2:
+                sys.stderr.write("  Running the embedder.\n\n")
+            physical.embedding = self._find_embedding(edges, hw_adj, max_qubits, **embed_args)
         else:
-            sys.stderr.write("  Running the embedder.\n\n")
-            physical.embedding = self._find_embedding(edges, hw_adj, verbose=1)
+            if verbosity >= 2:
+                sys.stderr.write("  Running the embedder, limiting it to %d qubits.\n\n" % max_qubits)
+            physical.embedding = self._impose_qubit_packing(max_qubits, edges, hw_adj, **embed_args)
+        if verbosity >= 2:
             sys.stderr.write("\n")
+        if physical.embedding == {}:
+            self.qmasm.abend("Failed to find an embedding")
 
         # Cache the embedding for next time.
         ec.write(physical.embedding)
@@ -331,11 +434,11 @@ class Sampler(object):
             sys.stderr.write("  Caching the embedding as %s.\n\n" % ec.hash)
         return physical
 
-    def embed_problem(self, logical, topology_file, verbosity):
+    def embed_problem(self, logical, topology_file, max_qubits, physical_nums, verbosity):
         "Embed a problem on a physical topology, if necessary."
         # Embed the problem.  We first filter out isolated variables that don't
         # appear in the embedding graph to prevent embed_bqm from complaining.
-        physical = self.find_problem_embedding(logical, topology_file, verbosity)
+        physical = self.find_problem_embedding(logical, topology_file, max_qubits, physical_nums, verbosity)
         if physical.embedding == {}:
             # No embedding is necessary.
             physical.logical = logical
@@ -346,8 +449,11 @@ class Sampler(object):
         for q, wt in physical.bqm.linear.items():
             if q not in physical.embedding and wt != 0.0:
                 self.qmasm.abend("Logical qubit %d has a nonzero weight (%.5g) but was not embedded" % (q, wt))
-        physical.bqm = embed_bqm(physical.bqm, physical.embedding,
-                                 physical.hw_adj, -physical.qmasm.chain_strength)
+        try:
+            physical.bqm = embed_bqm(physical.bqm, physical.embedding,
+                                     physical.hw_adj, -physical.qmasm.chain_strength)
+        except Exception as e:
+            self.qmasm.abend("Failed to embed the problem: %s" % e)
 
         # Update weights and strengths.  Maintain a reference to the
         # logical problem.
@@ -463,16 +569,11 @@ class Sampler(object):
         merged.info["timing"] = timing
         return merged
 
-    def _submit_and_block(self, bqm, verbosity, **params):
+    def _submit_and_block(self, sampler, bqm, store_params, **params):
         "Submit a job and wait for it to complete"
         # This method is a workaround for a bug in dwave-system.  See
         # https://github.com/dwavesystems/dwave-system/issues/297#issuecomment-632384524
-        sampler = self.sampler
         sub_params = copy.copy(params)
-        if self.qbsolv_sampler != None:
-            sub_params["solver"] = self.qbsolv_sampler
-            if verbosity >= 2:
-                sub_params["verbosity"] = 10   # Arbitrary large number
         result = None
         while result == None:
             # I don't know of a way to query a solver for the parameters it
@@ -480,62 +581,87 @@ class Sampler(object):
             # error string to determine what is and isn't acceptable.
             try:
                 result = sampler.sample(bqm, **sub_params)
+                if store_params:
+                    result.resolve()
+            except SolverFailureError as err:
+                match = self.unk_param_re.search(str(err))
+                if match == None:
+                    raise err
+                p = match[1]
+                with self.rejected_params_lock:
+                    self.rejected_params.append(p)
+                del sub_params[p]
+                result = None
             except TypeError as err:
                 match = self.unexp_arg_re.search(str(err))
                 if match == None:
-                    raise
+                    raise err
                 p = match[1]
-                self.rejected_params.append(p)
+                with self.rejected_params_lock:
+                    self.rejected_params.append(p)
                 del sub_params[p]
-        result.resolve()
+                result = None
+            except KeyError as err:
+                match = self.not_param_re.search(str(err))
+                if match == None:
+                    raise err
+                p = match[1]
+                with self.rejected_params_lock:
+                    self.rejected_params.append(p)
+                del sub_params[p]
+                result = None
+        if store_params:
+            self.final_params = sub_params
+            self.final_params_sem.release()
         return result
 
-    def acquire_samples(self, verbosity, physical, samples, anneal_time, spin_revs, postproc):
-        "Acquire a number of samples from either a hardware or software sampler."
-        # Map abbreviated to full names for postprocessing types.
-        postproc = {"none": "", "opt": "optimization", "sample": "sampling"}[postproc]
+    def _wrap_virtual_graph(self, sampler, bqm):
+        "Return a VirtualGraphComposite and an associated BQM."
+        # Explicitly scale the BQM to within the allowed h and J ranges.
+        bqm = bqm.copy()
+        props = sampler.properties
+        h_range = props["h_range"]
+        J_range = props["j_range"]
+        scale = min([h_range[0]/min(bqm.linear.values()),
+                     h_range[1]/max(bqm.linear.values()),
+                     J_range[0]/min(bqm.quadratic.values()),
+                     J_range[1]/max(bqm.quadratic.values())])
+        bqm.scale(scale)
 
-        # Determine the annealing time to use.
-        if anneal_time == None:
-            anneal_time = self._get_default_annealing_time()
+        # Define an identity mapping because the problem is already embedded.
+        id_embed = {q: [q] for q in bqm.linear.keys()}
+        id_embed.update({q[0]: [q[0]] for q in bqm.quadratic.keys()})
+        id_embed.update({q[1]: [q[1]] for q in bqm.quadratic.keys()})
 
-        # Compute a list of the number of samples to take each iteration
-        # and the number of spin reversals to perform.
-        samples_list = self._compute_sample_counts(samples, anneal_time)
-        spin_rev_list = self._compute_spin_rev_counts(spin_revs, samples_list)
-        nqmis = len(samples_list)   # Number of (non-unique) QMIs to submit
+        # Return a VirtualGraph and the new BQM.
+        return VirtualGraphComposite(sampler=self.sampler, embedding=id_embed), bqm
 
-        # QBSolv writes to standard output, but we want it to write to standard
-        # error instead.  Note that the following is imperfect because of C
-        # buffering.  Setting PYTHONUNBUFFERED=1 in the environment seems to
-        # help, though.
-        if self.qbsolv_sampler != None:
-            stdout_fileno = os.dup(sys.stdout.fileno())
-            os.dup2(sys.stderr.fileno(), sys.stdout.fileno())
-
-        # Submit all of our QMIs asynchronously.
-        results = [None for i in range(nqmis)]
-        executor = ThreadPoolExecutor()
-        overall_start_time = time.time_ns()
-        for i in range(nqmis):
-            solver_params = dict(chains=list(physical.embedding.values()),
-                                 num_reads=samples_list[i],
-                                 annealing_time=anneal_time,
-                                 num_spin_reversal_transforms=spin_rev_list[i],
-                                 postprocess=postproc)
-            for p in self.rejected_params:
-                del solver_params[p]
-            future = executor.submit(self._submit_and_block, physical.bqm, verbosity, **solver_params)
-            results[i] = SampleSet.from_future(future)
-
-        # Wait for the QMIs to finish.
-        executor.shutdown(wait=False)
+    def complete_sample_acquisition(self, verbosity, results, overall_start_time, physical):
+        "Wait for all results to complete then return a Solutions object."
+        nqmis = len(results)
         if verbosity == 1:
             if nqmis == 1:
                 sys.stderr.write("Waiting for the problem to complete.\n\n")
             else:
                 sys.stderr.write("Waiting for all %d subproblems to complete.\n\n" % nqmis)
         elif verbosity >= 2:
+            # Output our final parameters and their values.
+            sys.stderr.write("Parameters accepted by %s (first subproblem):\n\n" % self.client_info["solver_name"])
+            self.final_params_sem.acquire()
+            if len(self.final_params) == 0:
+                sys.stderr.write("    [none]\n")
+            else:
+                max_param_name_len = max([len(k) for k in self.final_params])
+                max_param_value_len = max([len(repr(v)) for v in self.final_params.values()])
+                sys.stderr.write("    %-*.*s  Value\n" %
+                                 (max_param_name_len, max_param_name_len, "Parameter"))
+                sys.stderr.write("    %s  %s\n" % ("-"*max_param_name_len, "-"*max_param_value_len))
+                for k, v in sorted(self.final_params.items()):
+                    sys.stderr.write("    %-*.*s  %s\n" %
+                                     (max_param_name_len, max_param_name_len, k, repr(v)))
+            sys.stderr.write("\n")
+
+            # Keep track of the number of subproblems completed.
             sys.stderr.write("Number of subproblems completed:\n\n")
             cdigits = len(str(nqmis))     # Digits in the number of completed QMIs
             tdigits = len(str(nqmis*5))   # Estimate 5 seconds per QMI submission
@@ -564,12 +690,88 @@ class Sampler(object):
                 sys.stderr.write("\n")
 
         # Merge the result of seperate runs into a composite answer.
-        answer = self._merge_results(results)
+        try:
+            answer = self._merge_results(results)
+        except SolverFailureError as err:
+            self.qmasm.abend("Solver error: %s" % err)
         answer.info["timing"]["round_trip_time"] = (overall_end_time - overall_start_time)//1000
 
         # Reset standard output.
-        if self.qbsolv_sampler != None:
+        if getattr(self, "qbsolv_sampler", None) != None:
             os.dup2(stdout_fileno, sys.stdout.fileno())
 
         # Return a Solutions object for further processing.
         return Solutions(answer, physical, verbosity >= 2)
+
+    def acquire_samples(self, verbosity, composites, physical, anneal_sched, samples, anneal_time, spin_revs, postproc):
+        "Acquire a number of samples from either a hardware or software sampler."
+        # Wrap composites around our sampler if requested.
+        sampler, bqm = self.sampler, physical.bqm
+        for c in composites:
+            try:
+                if c == "VirtualGraph":
+                    sampler, bqm = self._wrap_virtual_graph(sampler, bqm)
+                else:
+                    self.qmasm.abend('Internal error: unrecognized composite "%s"' % c)
+            except SystemExit as err:
+                raise err
+            except:
+                self.qmasm.abend("Failed to wrap a %s composite around the underlying sampler" % c)
+
+        # Map abbreviated to full names for postprocessing types.
+        postproc = {"none": "", "opt": "optimization", "sample": "sampling"}[postproc]
+
+        # Determine the annealing time to use.
+        if anneal_time == None:
+            anneal_time = self._get_default_annealing_time()
+
+        # Compute a list of the number of samples to take each iteration
+        # and the number of spin reversals to perform.
+        samples_list = self._compute_sample_counts(samples, anneal_time)
+        spin_rev_list = self._compute_spin_rev_counts(spin_revs, samples_list)
+        nqmis = len(samples_list)   # Number of (non-unique) QMIs to submit
+
+        # QBSolv writes to standard output, but we want it to write to standard
+        # error instead.  Note that the following is imperfect because of C
+        # buffering.  Setting PYTHONUNBUFFERED=1 in the environment seems to
+        # help, though.
+        if getattr(self, "qbsolv_sampler", None) != None:
+            stdout_fileno = os.dup(sys.stdout.fileno())
+            os.dup2(sys.stderr.fileno(), sys.stdout.fileno())
+
+        # Submit all of our QMIs asynchronously.
+        results = [None for i in range(nqmis)]
+        executor = ThreadPoolExecutor()
+        overall_start_time = time.time_ns()
+        for i in range(nqmis):
+            # Construct a set of solver parameters by combining typical
+            # parameters (e.g., num_reads) with solver-specific parameters
+            # (e.g., qpu_sampler), handling mutually exclusive parameters
+            # (e.g., annealing_time and anneal_schedule), and subtracting off
+            # any parameters previously rejected by the solver.
+            solver_params = dict(chains=list(physical.embedding.values()),
+                                 num_reads=samples_list[i],
+                                 num_spin_reversal_transforms=spin_rev_list[i],
+                                 postprocess=postproc)
+            solver_params.update(self.extra_solver_params)
+            if anneal_sched == None:
+                solver_params["annealing_time"] = anneal_time
+            else:
+                solver_params["anneal_schedule"] = anneal_sched
+            try:
+                # Some, but not all, solvers report the parameters they accept.
+                accepted_params = set(sampler.solver.properties["parameters"])
+                solver_params = {k: v for k, v in solver_params.items() if k in accepted_params}
+            except AttributeError:
+                pass
+            with self.rejected_params_lock:
+                for p in self.rejected_params:
+                    del solver_params[p]
+
+            # Submit the QMI to the solver in a background thread.
+            future = executor.submit(self._submit_and_block, sampler, bqm, i == 0, **solver_params)
+            results[i] = SampleSet.from_future(future)
+
+        # Wait for the QMIs to finish then return the results.
+        executor.shutdown(wait=False)
+        return self.complete_sample_acquisition(verbosity, results, overall_start_time, physical)
